__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


from tresor.gcell.FromSimulator import fromSimulator as scgmatsimu
from tresor.gsample.FromSimulator import fromSimulator as bulkgmatsimu
from tresor.util.file.Writer import Writer as pfwriter



def deepcvaesc_sc(

):
    from tresor.gcell.deepcvaesc.Predict import sample
    sample(model_fpn, num_labels=num_labels, batch_size=32)
    return


def spsimseq_sc(
        R_root=None,
        num_genes=10,
        num_cells=10,
        simulator='spsimseq',
        sv_fpn=False,
):
    gbycell, df_cells, df_genes = scgmatsimu(
        simulator=simulator,
        R_root=R_root,
        num_cells=num_genes,
        num_genes=num_cells,
    ).run()
    if sv_fpn:
        # pfwriter().generic(df=gbycell, sv_fpn=sv_fpn)
        gbycell.to_hdf(sv_fpn, key='df', mode='w')
    return gbycell, df_cells, df_genes


def spsimseq_bulk(
        R_root=None,
        num_samples=2,
        num_genes=10,
        simulator='spsimseq',
        sv_fpn=False,
):
    gspl = bulkgmatsimu(
        simulator=simulator,
        R_root=R_root,
        num_samples=num_samples,
        num_genes=num_genes,
    ).run()
    if sv_fpn:
        # pfwriter().generic(df=gspl, sv_fpn=sv_fpn)
        gspl.to_hdf(sv_fpn, key='df', mode='w')
    return gspl


def read(
        gmat_fpn,
):
    import pandas as pd
    df_gmat = pd.read_hdf(gmat_fpn, 'df')
    if 'Y' in df_gmat.columns:
        df_gmat = df_gmat.drop(columns=['Y'])
    return df_gmat


if __name__ == "__main__":
    from tresor.path import to

    # gbycell, _, _ = spsimseq_sc(
    #     R_root='D:/Programming/R/R-4.3.2/',
    #     num_genes=10,
    #     num_cells=10,
    #     simulator='spsimseq',
    #     sv_fpn=to('data/spsimseq_sc.h5'),
    # )
    # print(gbycell)

    # gspl = spsimseq_bulk(
    #     R_root='D:/Programming/R/R-4.3.2/',
    #     num_samples=2,
    #     num_genes=10,
    #     simulator='spsimseq',
    #     sv_fpn=to('data/spsimseq_bulk.h5'),
    # )
    # print(gspl)

    print(read(
        # gmat_fpn=to('data/gmat_customized.h5'),
        gmat_fpn=to('data/spsimseq_sc.h5'),
        # gmat_fpn=to('data/spsimseq_bulk.h5'),
    ))


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


import numpy as np
from tresor.util.plot.DimensionReduction import DimensionReduction as drplot


def sample(
        model_fpn,
        num_cells_per_cluster=100,
        image_size=144,
        num_labels=11,
        batch_size=32,
):
    import tensorflow as tf
    decoder = tf.keras.models.load_model(model_fpn)
    print(decoder)
    xs = []
    ys = []
    for ilabel in range(num_labels):
        print('===>label {}'.format(ilabel))
        digit_size = image_size
        z_combo_spl = np.random.normal(loc=1, scale=4, size=[num_cells_per_cluster, 2])
        # z_combo_spl = np.random.uniform(0, 4, size=[num_, 2])
        ilabel_ = np.eye(num_labels)[np.array([ilabel])]
        for i, zcspl in enumerate(z_combo_spl):
            z_sample = np.array([zcspl])
            x_decoded = decoder.get_layer("decoder").predict([z_sample, ilabel_])
            # print(x_decoded.shape)
            # print(x_decoded[0].shape)
            digit = x_decoded[0].reshape(digit_size, digit_size)
            # print(digit)
            xs.append(digit.reshape(digit_size * digit_size))
            t = digit.reshape(digit_size * digit_size)*260
            t = np.floor(t).astype(int)
            print(t[t>1])
            print(t[t>1].shape)
            # xs.append(x_decoded[0])
            ys.append(ilabel)
            # sds = digit * 255
            # print(np.where(sds > 1, sds, 0))
    drplot().single(X=xs, y=ys, tech='TSNE', marker_size=3, cmap='tab20b', title='CondiCVAE')
    drplot().single(X=xs, y=ys, tech='PCA', marker_size=3, cmap='tab20b', title='CondiCVAE')
    drplot().single(X=xs, y=ys, tech='UMAP', marker_size=3, cmap='tab20b', title='CondiCVAE')


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


from functools import wraps


class Library:

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, deal):
        manage = self.kwargs['method']
        @wraps(deal)
        def build(ph, *args, **kwargs):
            res = deal(ph, **kwargs)
            if kwargs['is_sv'] is True:
                if manage == 'default':
                    with open(kwargs['lib_fpn'], 'a') as file:
                        file.write(res + "\n")
                elif manage == 'separate':
                    with open(kwargs['lib_fpn'], 'a') as file:
                        file.write(kwargs['res'] + "\n")
            return res
        return build


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


import time
import numpy as np
import pandas as pd
from tresor.util.random.Number import Number as rannum
from tresor.util.sequence.symbol.Single import Single as dnasgl
from tresor.read.error.Library import Library as errlib
from tresor.util.Console import Console


class Error:

    def __init__(
            self,
            method,
    ):
        self.method = method
        self.errlib = errlib()
        self.console = Console()

    def __call__(self, deal):
        from functools import wraps
        @wraps(deal)
        def indexing(ph, *args, **kwargs):
            res2p = deal(ph, **kwargs)
            # print(func(res2p))

            if 'verbose' in kwargs['params'].keys():
                self.console.verbose = kwargs['params']['verbose']
            else:
                self.console.verbose = True

            self.console.print('======>Generate PCR errors...')

            if self.method == 'err1d':
                self.console.print('=========>The error 2D method is being used')
                func = self.table1D
            elif self.method == 'err2d':
                self.console.print('=========>The error 2D method is being used')
                func = self.table2D
            elif self.method == 'bftree':
                self.console.print('=========>The boolean flag tree method is being used')
                func = self.table_tree
            elif self.method == 'sptree':
                self.console.print('=========>The ShortPath method is being used')
                func = self.table_tree
            elif self.method == 'mutation_table_minimum':
                self.console.print('=========>The mutation_table_minimum method is being used')
                func = self.table_mutation_minimum
            elif self.method == 'mutation_table_complete':
                self.console.print('=========>The mutation_table_complete method is being used')
                func = self.table_mutation_complete
            else: # default
                self.console.print('=========>The ResimPy method is being used')
                func = self.table_tree
            return func(res2p)
        return indexing

    def table1D(
            self,
            res2p,
            kind='index_by_same_len',
    ):
        """
        ..  test:
            -----
            # dna = ['A', 'T', 'G', 'C']
            # dna.remove(pcr_err_base)
            # dna_map = self.todict(dna, reverse=True)
            # data_pcr.loc[pos_err[0], 'read'][pos_err[1]] = dna_map[pseudo_num]
            # pcr_err_base = data_pcr[pos_err[0]][0][pos_err[1]]

            # d = [
            #     [i, j] for i in ampl_read_len_df.index for j in np.arange(ampl_read_len_df[i])
            # ]
            # print(len(d))
            # ampl_read_len_list = np.arange(len(d))[:, np.newaxis]
            # map_tab = np.concatenate((ampl_read_len_list, d), axis=1)
            # print(self.tactic1(map_tab))

        ..  Method:
            -------
            =>string concat
                for pos_err, pseudo_num in zip(arr_err_pos, pseudo_nums):
                    tar_read = data_pcr.loc[pos_err[0], 'read']
                    pcr_err_base = tar_read[pos_err[1]]
                    dna_map = dnasgl().todict(
                        nucleotides=dnasgl().getEleTrimmed(
                            ele_loo=pcr_err_base,
                            universal=True,
                        ),
                        reverse=True,
                    )
                    # print('before', data_pcr.loc[pos_err[0], 'read'][pos_err[1]])
                    data_pcr.loc[pos_err[0], 'read'] = tar_read[:pos_err[1]] + dna_map[pseudo_num] + tar_read[pos_err[1]+1:]
        :param res2p:
        :return:
        """
        # print(res2p.keys())
        pcr_stime = time.time()
        data_pcr = pd.DataFrame(res2p['data_spl'], columns=['read', 'sam_id', 'source'])
        del res2p['data_spl']
        # print(len(data_pcr['read'][0]))
        # print(data_pcr.shape[0])
        res2p['recorder_pcr_read_num'].append(data_pcr.shape[0])
        self.console.print('======>{} reads to be amplified'.format(data_pcr.shape[0]))
        self.console.print('======>constructing the position table starts...')
        # print(data_pcr)
        pcr_postable_stime = time.time()
        if kind == 'index_by_same_len':
            seq_pos_ids, seq_ids = self.postableIndexBySameLen(seq_len=len(data_pcr['read'][0]), num_seq=data_pcr.shape[0])
        elif kind == 'index_by_lambda':
            seq_ids = []
            seq_pos_ids = []
            data_pcr.apply(lambda x: self.postableLambda(x, seq_ids, seq_pos_ids), axis=1)
        else:
            seq_pos_ids, seq_ids = self.postableIndexBySameLen(seq_len=len(data_pcr['read'][0]), num_seq=data_pcr.shape[0])
        # print(seq_ids)
        pos_table = {'seq_ids': seq_ids, 'seq_pos_ids': seq_pos_ids}
        self.console.print('======>time for constructing the position table  {time:.3f}s'.format(time=time.time() - pcr_postable_stime))
        ampl_nt_num = len(seq_ids)
        self.console.print('======>{} nucleotides to be amplified'.format(ampl_nt_num))
        res2p['recorder_nucleotide_num'].append(ampl_nt_num)
        self.console.print('======>determining PCR error numbers starts...')
        pcr_err_num_simu_stime = time.time()
        if res2p['err_num_met'] == 'binomial':
            pcr_err_num = rannum().binomial(n=ampl_nt_num, p=res2p['pcr_error'], use_seed=True, seed=res2p['ipcr'] + 1)
        elif res2p['err_num_met'] == 'nbinomial':
            pcr_err_num = rannum().nbinomial(
                n=ampl_nt_num*(1-res2p['pcr_error']),
                p=1-res2p['pcr_error'],
                use_seed=True,
                seed=res2p['ipcr'] + 1
            )
            # print(pcr_err_num)
        else:
            pcr_err_num = rannum().binomial(n=ampl_nt_num, p=res2p['pcr_error'], use_seed=True, seed=res2p['ipcr'] + 1)
        self.console.print('======>{} nucleotides to be erroneous at this PCR'.format(pcr_err_num))
        res2p['recorder_pcr_err_num'].append(pcr_err_num)
        spl_nt_ids = rannum().uniform(low=0, high=ampl_nt_num, num=pcr_err_num, use_seed=True, seed=res2p['ipcr'] + 1)
        arr_err_pos = []
        for i in spl_nt_ids:
            arr_err_pos.append([pos_table['seq_ids'][i], pos_table['seq_pos_ids'][i]])
        pseudo_nums = rannum().uniform(low=0, high=3, num=pcr_err_num, use_seed=False)
        # print(pseudo_nums)
        self.console.print('======>time for determining PCR error numbers  {time:.3f}s'.format(time=time.time() - pcr_err_num_simu_stime))
        self.console.print('======>assigning PCR errors starts...')
        pcr_err_assign_stime = time.time()
        data_pcr['read'] = data_pcr.apply(lambda x: list(x['read']), axis=1)
        # caser = self.todict5(arr_err_pos)
        # data_pcr['read'] = data_pcr.apply(lambda x: self.worktable(x, caser), axis=1)
        for pos_err, pseudo_num in zip(arr_err_pos, pseudo_nums):
            pcr_err_base = data_pcr.loc[pos_err[0], 'read'][pos_err[1]]
            dna_map = dnasgl().todict(
                nucleotides=dnasgl().getEleTrimmed(
                    ele_loo=pcr_err_base,
                    universal=True,
                ),
                reverse=True,
            )
            # print('before', data_pcr.loc[pos_err[0], 'read'][pos_err[1]])
            data_pcr.loc[pos_err[0], 'read'][pos_err[1]] = dna_map[pseudo_num]
            # print('after', data_pcr.loc[pos_err[0], 'read'][pos_err[1]])
        self.console.print('======>time for assigning PCR errors {time:.2f}s'.format(time=time.time() - pcr_err_assign_stime))
        self.console.print('======>merging the PCR duplicates and all previous sequences starts...')
        del arr_err_pos
        del spl_nt_ids
        pcr_merge_stime = time.time()
        data_pcr['read'] = data_pcr.apply(lambda x: ''.join(x['read']), axis=1)
        data_pcr['source'] = 'pcr-' + str(res2p['ipcr'] + 1)
        # print(data_pcr)
        # print(res2p)
        data_pcr['read'] = data_pcr['read'].apply(lambda x: self.errlib.insertion(read=x, ins_rate=res2p['pcr_ins_rate']))
        data_pcr['read'] = data_pcr['read'].apply(lambda x: self.errlib.deletion(read=x, del_rate=res2p['pcr_del_rate']))

        data_pcr = np.array(data_pcr)
        # print(data_pcr)
        # print(data_pcr.shape)

        res2p['data'] = np.concatenate((res2p['data'], data_pcr), axis=0)
        del data_pcr
        # print(res2p['recorder_pcr_err_num'])
        self.console.print('======>time for merging sequences {time:.2f}s'.format(time=time.time() - pcr_merge_stime))
        self.console.print('======>Summary report:')
        self.console.print('=========>PCR time: {time:.2f}s'.format(time=time.time() - pcr_stime))
        self.console.print('=========>the dimensions of the data: number of reads: {}'.format(res2p['data'].shape))
        self.console.print('=========>the number of reads at this PCR: {}, '.format(res2p['recorder_pcr_read_num']))
        self.console.print('=========>the number of nucleotides at this PCR: {}, '.format(res2p['recorder_nucleotide_num']))
        self.console.print('=========>the number of errors at this PCR: {}, '.format(res2p['recorder_pcr_err_num']))
        return res2p

    def table2D(self, res2p):
        pcr_stime = time.time()
        data_pcr = pd.DataFrame(res2p['data_spl'], columns=['read', 'sam_id', 'source'])
        data_pcr['read_len'] = data_pcr['read'].apply(lambda x: len(x))
        data_pcr['num_err_per_read'] = data_pcr['read_len'].apply(lambda x: rannum().binomial(
            n=x, p=res2p['pcr_error'], use_seed=False, seed=res2p['ipcr'] + 1
        ))
        data_pcr['pos_err_per_read'] = data_pcr.apply(lambda x: rannum().uniform(
            low=0, high=x['read_len'], num=x['num_err_per_read'], use_seed=False, seed=res2p['ipcr'] + 1
        ), axis=1)
        data_pcr['base_roll_per_read'] = data_pcr['num_err_per_read'].apply(lambda x: rannum().uniform(
            low=0, high=3, num=x, use_seed=False
        ))
        data_pcr['read_pcr'] = data_pcr.apply(lambda x: self.errlib.change(
            read=x['read'],
            pos_list=x['pos_err_per_read'],
            base_list=x['base_roll_per_read'],
        ), axis=1)
        # print(data_pcr[['read', 'read_pcr', 'pos_err_per_read']])
        del res2p['data_spl']
        pcr_merge_stime = time.time()
        data_pcr['sam_id'] = data_pcr['sam_id'].apply(lambda x: x + '_' + str(res2p['ipcr'] + 1))
        data_pcr['source'] = 'pcr-' + str(res2p['ipcr'] + 1)
        ### data_pcr.columns
        # ['read', 'sam_id', 'source', 'read_len', 'num_err_per_read',
        #         'pos_err_per_read', 'base_roll_per_read', 'read_pcr']
        ### data_pcr
        # read  ...                                           read_pcr
        # 0     CGCGTTAGTAATTCATTTTTTTCCCTTTAAAGGGTTTAAATTTGGG...  ...  CGCGTTAGTAATTCATTTTTTTCCCTTTAAAGGGTTTAAATTTGGG...
        # 1     CGCGTTAGTAATTCATCCCTTTAAAAAATTTGGGTTTGGGAAAAAA...  ...  CGCGTTAGTAATTCATCCCTTTAAAAAATTTGGGTTTGGGAAAAAA...
        # 2     CGCGTTAGTAATTCATGGGCCCTTTTTTAAATTTAAACCCCCCAAA...  ...  CGCGTTAGTAATTCATGGGCCCTTTTTTAAATTTAAACCCCCCAAA...
        # ...                                                 ...  ...                                                ...
        # 8692  GAAATCATGTAGTTCGCCCCCCAAATTTAAACCCAAATTTCCCAAA...  ...  GAAATCATGTAGTTCGCCCCCCAAATTTAAACCCAAATTTCCCAAA...
        # 8693  CGCGTTAGTAATTCATTTTGGGGGGAAACCCAAACCCCCCCCCGGG...  ...  CGCGTTAGTAATTCATTTTGGGGGGAAACCCAAACCCCCCCCCGGG...
        # [8694 rows x 8 columns]
        # print(data_pcr)
        # print(res2p)
        data_pcr['read'] = data_pcr['read'].apply(
            lambda x: self.errlib.insertion(read=x, ins_rate=res2p['pcr_ins_rate']))
        data_pcr['read'] = data_pcr['read'].apply(
            lambda x: self.errlib.deletion(read=x, del_rate=res2p['pcr_del_rate']))

        data_pcr = np.array(data_pcr[['read_pcr', 'sam_id', 'source']])
        res2p['data'] = np.concatenate((res2p['data'], data_pcr), axis=0)
        del data_pcr
        self.console.print('======>time for merging sequences {time:.2f}s'.format(time=time.time() - pcr_merge_stime))
        self.console.print('======>Summary report:')
        self.console.print('=========>PCR time: {time:.2f}s'.format(time=time.time() - pcr_stime))
        self.console.print('=========>the dimensions of the data: number of reads: {}'.format(res2p['data'].shape))
        self.console.print('=========>the number of reads at this PCR: {}, '.format(res2p['recorder_pcr_read_num']))
        self.console.print('=========>the number of nucleotides at this PCR: {}, '.format(res2p['recorder_nucleotide_num']))
        self.console.print('=========>the number of errors at this PCR: {}, '.format(res2p['recorder_pcr_err_num']))
        return res2p

    def table_mutation_minimum(self, res2p):
        """

        Notes
        -----

        Parameters
        ----------
        res2p

        Returns
        -------

        """
        pcr_stime = time.time()
        df_mut_info = pd.DataFrame()
        data_pcr = pd.DataFrame(res2p['data_spl'], columns=['read_len', 'sam_id', 'source'])
        data_pcr['num_err_per_read'] = data_pcr['read_len'].apply(lambda x: rannum().binomial(
            n=int(x), p=res2p['pcr_error'], use_seed=False, seed=res2p['ipcr'] + 1
        ))
        # data_pcr['num_err_per_read']
        # index num_err_per_read
        # 0     0
        # 1     0
        # 2     1
        # ...
        # 41    0
        # 42    0

        df_mut_info['pos_err_per_read'] = data_pcr.apply(lambda x: rannum().uniform(
            low=0, high=x['read_len'], num=x['num_err_per_read'], use_seed=False, seed=res2p['ipcr'] + 1
        ), axis=1)
        df_mut_info['base_roll_per_read'] = data_pcr['num_err_per_read'].apply(lambda x: rannum().uniform(
            low=0, high=3, num=x, use_seed=False
        ))

        # print(data_pcr[['read', 'read_pcr', 'pos_err_per_read']])
        del res2p['data_spl']
        pcr_merge_stime = time.time()
        data_pcr['sam_id'] = data_pcr['sam_id'].apply(lambda x: x + '_' + str(res2p['ipcr'] + 1))
        data_pcr['source'] = 'pcr-' + str(res2p['ipcr'] + 1)

        # print(df_mut_info)
        df_mut_info['sam_id'] = data_pcr['sam_id'].copy()
        # print(data_pcr)
        df_mut_info['mark'] = df_mut_info['pos_err_per_read'].apply(lambda x: 1 if x.size == 0 else 0)
        # print(df_mut_info)
        df_mut_info = df_mut_info.drop(df_mut_info.loc[df_mut_info['mark'] == 1].index)
        df_mut_info = df_mut_info.drop(['mark'], axis=1)
        ### df_mut_info
        # pos_err_per_read base_roll_per_read                          sam_id
        # 201                 [9]                [1]           154*c*1*g*4*_3_4_9_10
        # 264                [47]                [1]        53*c*1*g*4*_1_2_6_7_9_10
        # 870                [28]                [0]           7*c*1*g*3*_3_4_8_9_10
        # 946                [13]                [2]        13*c*1*g*0*_1_3_4_8_9_10
        # 1312               [40]                [1]      68*c*1*g*4*_3_5_6_7_8_9_10
        # ...                 ...                ...                             ...
        # 152258             [48]                [2]             115*c*1*g*4*_4_5_10
        # 152548              [6]                [1]         164*c*1*g*4*_3_4_6_7_10
        # 152656             [30]                [2]         173*c*1*g*4*_1_5_7_9_10
        # 152914             [33]                [0]  46*c*1*g*4*_1_2_3_4_5_6_7_9_10
        # 153185              [5]                [2]         177*c*1*g*4*_2_5_6_8_10
        #
        # [792 rows x 3 columns]
        data_pcr = np.array(data_pcr[['read_len', 'sam_id', 'source']])
        res2p['data'] = np.concatenate((res2p['data'], data_pcr), axis=0)
        # print(np.concatenate((res2p['mut_info'], np.array(df_mut_info)), axis=0))
        res2p['mut_info'] = np.concatenate((res2p['mut_info'], np.array(df_mut_info)), axis=0)
        del data_pcr
        self.console.print('======>Time for merging sequences {time:.2f}s'.format(time=time.time() - pcr_merge_stime))
        self.console.print('======>Summary report:')
        self.console.print('=========>PCR time: {time:.2f}s'.format(time=time.time() - pcr_stime))
        self.console.print('=========>Dimension of the data: number of reads: {}'.format(res2p['data'].shape))
        self.console.print('=========>Number of reads at the PCR: {}, '.format(res2p['recorder_pcr_read_num']))
        self.console.print('=========>Number of nucleotides at the PCR: {}, '.format(res2p['recorder_nucleotide_num']))
        self.console.print('=========>Number of errors at the PCR: {}, '.format(res2p['recorder_pcr_err_num']))
        return res2p

    def table_mutation_complete(self, res2p):
        """

        Notes
        -----

        Parameters
        ----------
        res2p

        Returns
        -------

        """
        pcr_stime = time.time()
        df_mut_info = pd.DataFrame()
        # print(res2p)
        # print(res2p['data_spl'])

        data_pcr = pd.DataFrame(res2p['data_spl'], columns=[
            'read_len',
            'sam_id',
            'source',
            'bead_mut',
            'bead_del',
            'bead_ins',
            'pcr_err_mark',
        ])
        # print(data_pcr)
        data_pcr['num_err_per_read'] = data_pcr['read_len'].apply(lambda x: rannum().binomial(
            n=int(x), p=res2p['pcr_error'], use_seed=False, seed=res2p['ipcr'] + 1
        ))
        data_pcr['pcr_err_mark'] = data_pcr['num_err_per_read'].apply(lambda x: False if x == 0 else True)
        # print(data_pcr['pcr_err_mark'])
        # @@ data_pcr['num_err_per_read']
        # index num_err_per_read
        # 0     0
        # 1     0
        # 2     1
        # ...
        # 41    0
        # 42    0

        df_mut_info['pos_err_per_read'] = data_pcr.apply(lambda x: rannum().uniform(
            low=0, high=x['read_len'], num=x['num_err_per_read'], use_seed=False, seed=res2p['ipcr'] + 1
        ), axis=1)
        df_mut_info['base_roll_per_read'] = data_pcr['num_err_per_read'].apply(lambda x: rannum().uniform(
            low=0, high=3, num=x, use_seed=False
        ))

        # print(data_pcr[['read', 'read_pcr', 'pos_err_per_read']])
        del res2p['data_spl']
        pcr_merge_stime = time.time()
        data_pcr['sam_id'] = data_pcr['sam_id'].apply(lambda x: x + '_' + str(res2p['ipcr'] + 1))
        data_pcr['source'] = 'pcr-' + str(res2p['ipcr'] + 1)

        ### df_mut_info
        # pos_err_per_read base_roll_per_read
        # 0                    []                 []
        # 1                    []                 []
        # 2                    []                 []
        # 3                    []                 []
        # 4                    []                 []
        # ...                 ...                ...
        # 153595               []                 []
        # 153596               []                 []
        # 153597               []                 []
        # 153598               []                 []
        # 153599               []                 []
        df_mut_info['sam_id'] = data_pcr['sam_id'].copy()
        df_mut_info['bead_mut'] = data_pcr['bead_mut'].copy()
        df_mut_info['bead_del'] = data_pcr['bead_del'].copy()
        df_mut_info['bead_ins'] = data_pcr['bead_ins'].copy()
        df_mut_info['pcr_err_mark'] = data_pcr['pcr_err_mark'].copy()

        data_pcr = np.array(data_pcr[[
            'read_len',
            'sam_id',
            'source',
            'bead_mut',
            'bead_del',
            'bead_ins',
            'pcr_err_mark',
        ]])
        # print(data_pcr)
        # print(res2p['data'])

        # print(res2p['mut_info'])
        # print(df_mut_info)
        res2p['data'] = np.concatenate((res2p['data'], data_pcr), axis=0)
        # print(np.concatenate((res2p['mut_info'], np.array(df_mut_info)), axis=0))
        res2p['mut_info'] = np.concatenate((res2p['mut_info'], np.array(df_mut_info)), axis=0)
        self.console.print('======>Time for merging sequences {time:.2f}s'.format(time=time.time() - pcr_merge_stime))
        self.console.print('======>Summary report:')
        self.console.print('=========>PCR time: {time:.2f}s'.format(time=time.time() - pcr_stime))
        self.console.print('=========>Dimension of the data: number of reads: {}'.format(res2p['data'].shape))
        self.console.print('=========>Number of reads at the PCR: {}, '.format(res2p['recorder_pcr_read_num']))
        self.console.print('=========>Number of nucleotides at the PCR: {}, '.format(res2p['recorder_nucleotide_num']))
        self.console.print('=========>Number of errors at the PCR: {}, '.format(res2p['recorder_pcr_err_num']))
        del data_pcr
        return res2p

    def table_tree(self, res2p):
        pcr_stime = time.time()
        data_pcr = pd.DataFrame(
            res2p['data_spl'],
            columns=[
                'sam_id',
                'source',
            ]
        )
        del res2p['data_spl']
        data_pcr['sam_id'] = data_pcr['sam_id'].apply(lambda x: x + '_' + str(res2p['ipcr'] + 1))
        data_pcr['source'] = 'pcr-' + str(res2p['ipcr'] + 1)
        # print(data_pcr)
        data_pcr = np.array(data_pcr)
        res2p['data'] = np.concatenate((res2p['data'], data_pcr), axis=0)
        # print(res2p['data'])
        # print(res2p['data'].shape)
        del data_pcr
        self.console.print('======>Summary report:')
        self.console.print('=========>Number of reads at this PCR: {}'.format(res2p['data'].shape))
        self.console.print('=========>Time for PCR tree construction: {time:.2f}s'.format(time=time.time() - pcr_stime))
        return res2p

    def postableIndexBySameLen(self, seq_len, num_seq):
        """
        If each read has 10 positions, this function returns
        seq_ids
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, ...
        seq_pos_ids
        0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, ...

        Parameters
        ----------
        seq_len
        num_seq

        Returns
        -------
        seq_ids
            1d list, the id of the ith read
        seq_pos_ids
            1d list, the base positions (starting from 0 every time for each read) of reads

        """
        nt_ids = [i for i in range(seq_len)]
        seq_pos_ids = nt_ids * num_seq
        seq_ids = np.array([[i] * seq_len for i in range(num_seq)]).ravel().tolist()
        return seq_pos_ids, seq_ids

    def postableLambda(self, x, ids, pos_ids):
        for i in range(x['read_len']):
            pos_ids.append(i)
        ids = ids + [x.name] * x['read_len']
        return ids, pos_ids

    def worktable(self, x, ids):
        if x.name not in ids:
            return x['read']
        else:
            tt = x['read']
            for i in ids[x.name]:
                pcr_err_base = tt[i]
                dna_map = dnasgl().todict(
                    nucleotides=dnasgl().getEleTrimmed(
                        ele_loo=pcr_err_base,
                        universal=True,
                    ),
                    reverse=True,
                )
                tt = tt[:i] + dna_map[2] + tt[i+1:]
            return tt

    def postable_deprecate(self, x, ids, pos_ids):
        for i in range(x['read_len']):
            pos_ids.append(i)
        for i in [x.name] * x['read_len']:
            ids.append(i)
        return ids, pos_ids

    def epos_deprecate(self, spl_nt_ids, ampl_read_len_list):
        """
        ..  @call:
            ------
            arr_err_pos = self.epos(spl_nt_ids=spl_nt_ids, ampl_read_len_list=ampl_read_len_list)

        :param spl_nt_ids: is a list of ids of nucleotides to be modified.
        :param ampl_read_len_list: a list of lengths of reads to be amplified.
        :return:
        """
        pos_err = []
        for i in spl_nt_ids:
            cc = 0
            for id, j in enumerate(ampl_read_len_list):
                cc += j
                if cc >= i:
                    # print(cc, i)
                    pos_err.append([id, (j - 1) - (cc % i)])
                    break
        return pos_err

    def epos2_deprecate(self, ampl_read_len_list):
        ids = []
        pos = []
        for k, e in enumerate(ampl_read_len_list):
            l = [k] * e
            ids = ids + l
            p = [i for i in range(e)]
            pos = pos + p
        return np.array([pos, ids])

    def tactic1(self, arr_2d):
        result = {}
        len_arr = len(arr_2d[0])
        if len_arr == 2:
            for item in arr_2d:
                result[item[0]] = item[1]
        else:
            for item in arr_2d:
                result[item[0]] = item[1:]
        return result

    def todict(self, nucleotides, reverse=False):
        aa_dict = {}
        for k, v in enumerate(nucleotides):
            aa_dict[v] = k
        if reverse:
            aa_dict = {v: k for k, v in aa_dict.items()}
        return aa_dict

    def todict5(self, arr_2d):
        """
        ..  @description:
            -------------
            2d arr to 1d dict, each key in the 1d dict being an list

        :param arr_2d:
        :return:
        """
        result = {}
        for item in arr_2d:
            result[item[0]] = []
        for item in arr_2d:
            if item[0] in result.keys():
                result[item[0]].append(item[1])
                # print(result[item[0]])
        return result


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


import time
from tresor.library.SingleLocus import SingleLocus as libslocus
from tresor.scenario.seqerr.SingleLocus import SingleLocus as seqerr
from tresor.scenario.pcrerr.SingleLocus import SingleLocus as pcrerr
from tresor.scenario.pcrnum.SingleLocus import SingleLocus as pcrnum
from tresor.scenario.amplrate.SingleLocus import SingleLocus as amplrate
from tresor.scenario.umilen.SingleLocus import SingleLocus as umilen
from tresor.scenario.seqdep.SingleLocus import SingleLocus as seqdep
from tresor.scenario.generic.SingleLocus import SingleLocus as generic


def library(
        working_dir,
        seq_num,
        sim_thres,
        permutation,

        is_seed=True,
        is_sv_umi_lib=True,
        is_sv_seq_lib=True,
        is_sv_primer_lib=True,
        is_sv_adapter_lib=True,
        is_sv_spacer_lib=True,

        bead_mutation=False,
        bead_mut_rate=1e-4,
        bead_deletion=False,
        bead_del_rate=0.016,
        bead_insertion=False,
        bead_ins_rate=7.1e-7,

        len_params=None,
        seq_params=None,
        material_params=None,
        condis=None,
        mode=None,

        config_fpn=None,
        verbose=True,
):
    if config_fpn:
        import yaml
        with open(config_fpn, "r") as f:
            configs = yaml.safe_load(f)
            # for k, item in configs.items():
            #     print(k, item)
        len_params=configs['len_params']
        seq_params=configs['seq_params']
        material_params=configs['material_params']
        condis=configs['condis']

    libslocus(
        seq_num=seq_num,
        len_params=len_params,
        seq_params=seq_params,
        material_params=material_params,
        condis=condis,

        working_dir=working_dir,

        sim_thres=sim_thres,
        permutation=permutation,

        mode=mode,  # long_read short_read

        is_sv_umi_lib=is_sv_umi_lib,
        is_sv_seq_lib=is_sv_seq_lib,
        is_sv_primer_lib=is_sv_primer_lib,
        is_sv_adapter_lib=is_sv_adapter_lib,
        is_sv_spacer_lib=is_sv_spacer_lib,

        bead_mutation=bead_mutation,
        bead_mut_rate=bead_mut_rate,
        bead_deletion=bead_deletion,
        bead_del_rate=bead_del_rate,
        bead_insertion=bead_insertion,
        bead_ins_rate=bead_ins_rate,

        is_seed=is_seed,
        verbose=verbose,  # False True
    ).pooling()
    return 'Finished'


def simu_seq_err(
        seq_num,
        working_dir,
        sim_thres,
        permutation,

        len_params=None,
        condis=None,

        ampl_rate=None,
        err_route=None,
        pcr_error=None,
        pcr_num=None,
        err_num_met=None,
        seq_errors=None,
        seq_sub_spl_number=None,
        seq_sub_spl_rate=None,

        use_seed=True,
        seed=1,

        pcr_deletion=False,
        pcr_insertion=False,
        pcr_del_rate=0,
        pcr_ins_rate=0,
        seq_deletion=False,
        seq_insertion=False,
        seq_del_rate=0,
        seq_ins_rate=0,

        is_sv_umi_lib=True,
        is_sv_seq_lib=True,
        is_sv_primer_lib=True,
        is_sv_adapter_lib=True,
        is_sv_spacer_lib=True,
        sv_fastq_fp=True,

        config_fpn=None,
        verbose=True,

        **kwargs,
):
    if config_fpn:
        import yaml
        with open(config_fpn, "r") as f:
            configs = yaml.safe_load(f)
            # for k, item in configs.items():
            #     print(k, item)
        len_params=configs['len_params']
        kwargs['seq_params']=configs['seq_params']
        kwargs['material_params']=configs['material_params']
        condis=configs['condis']

        ampl_rate = configs['ampl_rate']
        err_route = configs['err_route']
        pcr_error = configs['pcr_error']
        pcr_num = configs['pcr_num']
        err_num_met = configs['err_num_met']
        seq_errors = configs['seq_errors']
        seq_sub_spl_number = configs['seq_sub_spl_number']
        seq_sub_spl_rate = configs['seq_sub_spl_rate']
        if "pcr_deletion" in configs.keys():
            pcr_deletion = configs['pcr_deletion']
        if "pcr_insertion" in configs.keys():
            pcr_insertion = configs['pcr_insertion']
        if "pcr_del_rate" in configs.keys():
            pcr_del_rate = configs['pcr_del_rate']
        if "pcr_ins_rate" in configs.keys():
            pcr_ins_rate = configs['pcr_ins_rate']
        if "seq_deletion" in configs.keys():
            seq_deletion = configs['seq_deletion']
        if "seq_insertion" in configs.keys():
            seq_insertion = configs['seq_insertion']
        if "seq_del_rate" in configs.keys():
            seq_del_rate = configs['seq_del_rate']
        if "seq_ins_rate" in configs.keys():
            seq_ins_rate = configs['seq_ins_rate']

    stime = time.time()
    res_dict = seqerr(
        # initial sequence generation
        len_params=len_params,
        mode=kwargs['mode'] if 'mode' in kwargs.keys() else None,
        material_params=kwargs['material_params'] if 'material_params' in kwargs.keys() else None,
        seq_params=kwargs['seq_params'] if 'seq_params' in kwargs.keys() else None,
        seq_num=seq_num,
        working_dir=working_dir,

        is_sv_umi_lib=is_sv_umi_lib,
        is_sv_seq_lib=is_sv_seq_lib,
        is_sv_primer_lib=is_sv_primer_lib,
        is_sv_adapter_lib=is_sv_adapter_lib,
        is_sv_spacer_lib=is_sv_spacer_lib,
        condis=condis,
        sim_thres=sim_thres,
        permutation=permutation,

        # PCR amplification
        ampl_rate=ampl_rate,
        err_route=err_route,
        pcr_error=pcr_error,
        pcr_num=pcr_num,
        err_num_met=err_num_met,
        seq_errors=seq_errors,
        seq_sub_spl_number=seq_sub_spl_number,
        seq_sub_spl_rate=seq_sub_spl_rate,

        # indels
        pcr_deletion=pcr_deletion,
        pcr_insertion=pcr_insertion,
        pcr_del_rate=pcr_del_rate,
        pcr_ins_rate=pcr_ins_rate,
        seq_deletion=seq_deletion,
        seq_insertion=seq_insertion,
        seq_del_rate=seq_del_rate,
        seq_ins_rate=seq_ins_rate,

        use_seed=use_seed,
        seed=seed,

        verbose=verbose,

        sv_fastq_fp=sv_fastq_fp,
    ).generate()
    etime = time.time()
    print("===>Time: {:.3f}s".format(etime - stime))
    print('Finished!')
    return res_dict


def simu_pcr_err(
        seq_num,
        working_dir,
        sim_thres,
        permutation,

        len_params=None,
        condis=None,
        ampl_rate=None,
        err_route=None,
        pcr_errors=None,
        seq_error=None,
        pcr_num=None,
        err_num_met=None,

        use_seed=True,
        seed=1,

        pcr_deletion=False,
        pcr_insertion=False,
        pcr_del_rate=0,
        pcr_ins_rate=0,
        seq_deletion=False,
        seq_insertion=False,
        seq_del_rate=0,
        seq_ins_rate=0,

        is_sv_umi_lib=True,
        is_sv_seq_lib=True,
        is_sv_primer_lib=True,
        is_sv_adapter_lib=True,
        is_sv_spacer_lib=True,
        sv_fastq_fp=True,

        seq_sub_spl_number=None,
        seq_sub_spl_rate=1/3,

        config_fpn=None,

        verbose=True,
        **kwargs,
):
    if config_fpn:
        import yaml
        with open(config_fpn, "r") as f:
            configs = yaml.safe_load(f)
            # for k, item in configs.items():
            #     print(k, item)
        len_params=configs['len_params']
        kwargs['seq_params']=configs['seq_params']
        kwargs['material_params']=configs['material_params']
        condis=configs['condis']

        ampl_rate = configs['ampl_rate']
        err_route = configs['err_route']
        seq_error = configs['seq_error']
        pcr_num = configs['pcr_num']
        err_num_met = configs['err_num_met']
        pcr_errors = configs['pcr_errors']
        seq_sub_spl_number = configs['seq_sub_spl_number']
        seq_sub_spl_rate = configs['seq_sub_spl_rate']
        if "pcr_deletion" in configs.keys():
            pcr_deletion = configs['pcr_deletion']
        if "pcr_insertion" in configs.keys():
            pcr_insertion = configs['pcr_insertion']
        if "pcr_del_rate" in configs.keys():
            pcr_del_rate = configs['pcr_del_rate']
        if "pcr_ins_rate" in configs.keys():
            pcr_ins_rate = configs['pcr_ins_rate']
        if "seq_deletion" in configs.keys():
            seq_deletion = configs['seq_deletion']
        if "seq_insertion" in configs.keys():
            seq_insertion = configs['seq_insertion']
        if "seq_del_rate" in configs.keys():
            seq_del_rate = configs['seq_del_rate']
        if "seq_ins_rate" in configs.keys():
            seq_ins_rate = configs['seq_ins_rate']

    stime = time.time()
    res_dict = pcrerr(
        len_params=len_params,
        mode=kwargs['mode'] if 'mode' in kwargs.keys() else None,
        material_params=kwargs['material_params'] if 'material_params' in kwargs.keys() else None,
        seq_params=kwargs['seq_params'] if 'seq_params' in kwargs.keys() else None,
        seq_num=seq_num,
        is_sv_umi_lib=is_sv_umi_lib,
        is_sv_seq_lib=is_sv_seq_lib,
        is_sv_primer_lib=is_sv_primer_lib,
        is_sv_adapter_lib=is_sv_adapter_lib,
        is_sv_spacer_lib=is_sv_spacer_lib,
        working_dir=working_dir,
        condis=condis,
        sim_thres=sim_thres,
        permutation=permutation,
        err_route=err_route,
        pcr_errors=pcr_errors,
        seq_error=seq_error,
        pcr_num=pcr_num,
        err_num_met=err_num_met,

        # indels
        pcr_deletion=pcr_deletion,
        pcr_insertion=pcr_insertion,
        pcr_del_rate=pcr_del_rate,
        pcr_ins_rate=pcr_ins_rate,
        seq_deletion=seq_deletion,
        seq_insertion=seq_insertion,
        seq_del_rate=seq_del_rate,
        seq_ins_rate=seq_ins_rate,

        use_seed=use_seed,
        seed=seed,
        ampl_rate=ampl_rate,
        sv_fastq_fp=sv_fastq_fp,
        seq_sub_spl_number=seq_sub_spl_number,
        seq_sub_spl_rate=seq_sub_spl_rate,
        verbose=verbose,
    ).generate()
    etime = time.time()
    print("===>Time: {:.3f}s".format(etime - stime))
    print('Finished!')
    return res_dict


def simu_pcr_num(
        seq_num,
        working_dir,
        sim_thres,
        permutation,

        len_params=None,
        condis=None,
        ampl_rate=None,
        err_route=None,
        pcr_error=None,
        seq_error=None,
        pcr_nums=None,
        err_num_met=None,

        use_seed=True,
        seed=1,

        pcr_deletion=False,
        pcr_insertion=False,
        pcr_del_rate=0,
        pcr_ins_rate=0,
        seq_deletion=False,
        seq_insertion=False,
        seq_del_rate=0,
        seq_ins_rate=0,

        is_sv_umi_lib=True,
        is_sv_seq_lib=True,
        is_sv_primer_lib=True,
        is_sv_adapter_lib=True,
        is_sv_spacer_lib=True,
        sv_fastq_fp=True,

        seq_sub_spl_number=None,
        seq_sub_spl_rate=1 / 3,

        config_fpn=None,

        verbose=True,
        **kwargs,
):
    if config_fpn:
        import yaml
        with open(config_fpn, "r") as f:
            configs = yaml.safe_load(f)
            # for k, item in configs.items():
            #     print(k, item)
        len_params = configs['len_params']
        kwargs['seq_params'] = configs['seq_params']
        kwargs['material_params'] = configs['material_params']
        condis = configs['condis']

        ampl_rate = configs['ampl_rate']
        err_route = configs['err_route']
        seq_error = configs['seq_error']
        pcr_nums = configs['pcr_nums']
        err_num_met = configs['err_num_met']
        pcr_error = configs['pcr_error']
        seq_sub_spl_number = configs['seq_sub_spl_number']
        seq_sub_spl_rate = configs['seq_sub_spl_rate']
        if "pcr_deletion" in configs.keys():
            pcr_deletion = configs['pcr_deletion']
        if "pcr_insertion" in configs.keys():
            pcr_insertion = configs['pcr_insertion']
        if "pcr_del_rate" in configs.keys():
            pcr_del_rate = configs['pcr_del_rate']
        if "pcr_ins_rate" in configs.keys():
            pcr_ins_rate = configs['pcr_ins_rate']
        if "seq_deletion" in configs.keys():
            seq_deletion = configs['seq_deletion']
        if "seq_insertion" in configs.keys():
            seq_insertion = configs['seq_insertion']
        if "seq_del_rate" in configs.keys():
            seq_del_rate = configs['seq_del_rate']
        if "seq_ins_rate" in configs.keys():
            seq_ins_rate = configs['seq_ins_rate']

    stime = time.time()
    res_dict = pcrnum(
        len_params=len_params,
        mode=kwargs['mode'] if 'mode' in kwargs.keys() else None,
        material_params=kwargs['material_params'] if 'material_params' in kwargs.keys() else None,
        seq_params=kwargs['seq_params'] if 'seq_params' in kwargs.keys() else None,
        seq_num=seq_num,
        is_sv_umi_lib=is_sv_umi_lib,
        is_sv_seq_lib=is_sv_seq_lib,
        is_sv_primer_lib=is_sv_primer_lib,
        is_sv_adapter_lib=is_sv_adapter_lib,
        is_sv_spacer_lib=is_sv_spacer_lib,
        working_dir=working_dir,
        condis=condis,
        sim_thres=sim_thres,
        permutation=permutation,
        err_route=err_route,
        pcr_error=pcr_error,
        seq_error=seq_error,
        pcr_nums=pcr_nums,
        err_num_met=err_num_met,

        # indels
        pcr_deletion=pcr_deletion,
        pcr_insertion=pcr_insertion,
        pcr_del_rate=pcr_del_rate,
        pcr_ins_rate=pcr_ins_rate,
        seq_deletion=seq_deletion,
        seq_insertion=seq_insertion,
        seq_del_rate=seq_del_rate,
        seq_ins_rate=seq_ins_rate,

        use_seed=use_seed,
        seed=seed,
        ampl_rate=ampl_rate,
        sv_fastq_fp=sv_fastq_fp,
        seq_sub_spl_number=seq_sub_spl_number,
        seq_sub_spl_rate=seq_sub_spl_rate,
        verbose=verbose,
    ).generate()
    etime = time.time()
    print("===>Time: {:.3f}s".format(etime - stime))
    print('Finished!')
    return res_dict


def simu_ampl_rate(
        seq_num,
        working_dir,
        sim_thres,
        permutation,

        len_params=None,
        condis=None,
        ampl_rates=None,
        err_route=None,
        pcr_error=None,
        seq_error=None,
        pcr_num=None,
        err_num_met=None,

        use_seed=True,
        seed=1,

        pcr_deletion=False,
        pcr_insertion=False,
        pcr_del_rate=0,
        pcr_ins_rate=0,
        seq_deletion=False,
        seq_insertion=False,
        seq_del_rate=0,
        seq_ins_rate=0,

        is_sv_umi_lib=True,
        is_sv_seq_lib=True,
        is_sv_primer_lib=True,
        is_sv_adapter_lib=True,
        is_sv_spacer_lib=True,
        sv_fastq_fp=True,

        seq_sub_spl_number=None,
        seq_sub_spl_rate=1 / 3,

        config_fpn=None,

        verbose=True,
        **kwargs,
):
    if config_fpn:
        import yaml
        with open(config_fpn, "r") as f:
            configs = yaml.safe_load(f)
            # for k, item in configs.items():
            #     print(k, item)
        len_params = configs['len_params']
        kwargs['seq_params'] = configs['seq_params']
        kwargs['material_params'] = configs['material_params']
        condis = configs['condis']

        ampl_rates = configs['ampl_rates']
        err_route = configs['err_route']
        seq_error = configs['seq_error']
        pcr_num = configs['pcr_num']
        err_num_met = configs['err_num_met']
        pcr_error = configs['pcr_error']
        seq_sub_spl_number = configs['seq_sub_spl_number']
        seq_sub_spl_rate = configs['seq_sub_spl_rate']
        if "pcr_deletion" in configs.keys():
            pcr_deletion = configs['pcr_deletion']
        if "pcr_insertion" in configs.keys():
            pcr_insertion = configs['pcr_insertion']
        if "pcr_del_rate" in configs.keys():
            pcr_del_rate = configs['pcr_del_rate']
        if "pcr_ins_rate" in configs.keys():
            pcr_ins_rate = configs['pcr_ins_rate']
        if "seq_deletion" in configs.keys():
            seq_deletion = configs['seq_deletion']
        if "seq_insertion" in configs.keys():
            seq_insertion = configs['seq_insertion']
        if "seq_del_rate" in configs.keys():
            seq_del_rate = configs['seq_del_rate']
        if "seq_ins_rate" in configs.keys():
            seq_ins_rate = configs['seq_ins_rate']

    stime = time.time()
    res_dict = amplrate(
        len_params=len_params,
        mode=kwargs['mode'] if 'mode' in kwargs.keys() else None,
        material_params=kwargs['material_params'] if 'material_params' in kwargs.keys() else None,
        seq_params=kwargs['seq_params'] if 'seq_params' in kwargs.keys() else None,
        seq_num=seq_num,
        is_sv_umi_lib=is_sv_umi_lib,
        is_sv_seq_lib=is_sv_seq_lib,
        is_sv_primer_lib=is_sv_primer_lib,
        is_sv_adapter_lib=is_sv_adapter_lib,
        is_sv_spacer_lib=is_sv_spacer_lib,
        working_dir=working_dir,
        condis=condis,
        sim_thres=sim_thres,
        permutation=permutation,
        err_route=err_route,
        pcr_error=pcr_error,
        seq_error=seq_error,
        pcr_num=pcr_num,
        err_num_met=err_num_met,

        # indels
        pcr_deletion=pcr_deletion,
        pcr_insertion=pcr_insertion,
        pcr_del_rate=pcr_del_rate,
        pcr_ins_rate=pcr_ins_rate,
        seq_deletion=seq_deletion,
        seq_insertion=seq_insertion,
        seq_del_rate=seq_del_rate,
        seq_ins_rate=seq_ins_rate,

        use_seed=use_seed,
        seed=seed,
        ampl_rates=ampl_rates,
        sv_fastq_fp=sv_fastq_fp,
        seq_sub_spl_number=seq_sub_spl_number,
        seq_sub_spl_rate=seq_sub_spl_rate,
        verbose=verbose,
    ).generate()
    etime = time.time()
    print("===>Time: {:.3f}s".format(etime - stime))
    print('Finished!')
    return res_dict


def simu_umi_len(
        seq_num,
        working_dir,
        sim_thres,
        permutation,

        len_params=None,
        condis=None,
        ampl_rate=None,
        err_route=None,
        pcr_error=None,
        seq_error=None,
        pcr_num=None,
        err_num_met=None,

        use_seed=True,
        seed=1,

        pcr_deletion=False,
        pcr_insertion=False,
        pcr_del_rate=0,
        pcr_ins_rate=0,
        seq_deletion=False,
        seq_insertion=False,
        seq_del_rate=0,
        seq_ins_rate=0,

        is_sv_umi_lib=True,
        is_sv_seq_lib=True,
        is_sv_primer_lib=True,
        is_sv_adapter_lib=True,
        is_sv_spacer_lib=True,
        sv_fastq_fp=True,

        seq_sub_spl_number=None,
        seq_sub_spl_rate=1 / 3,

        config_fpn=None,

        verbose=True,
        **kwargs,
):
    if config_fpn:
        import yaml
        with open(config_fpn, "r") as f:
            configs = yaml.safe_load(f)
            # for k, item in configs.items():
            #     print(k, item)
        len_params = configs['len_params']
        kwargs['seq_params'] = configs['seq_params']
        kwargs['material_params'] = configs['material_params']
        condis = configs['condis']

        ampl_rate = configs['ampl_rate']
        err_route = configs['err_route']
        seq_error = configs['seq_error']
        pcr_num = configs['pcr_num']
        err_num_met = configs['err_num_met']
        pcr_error = configs['pcr_error']
        seq_sub_spl_number = configs['seq_sub_spl_number']
        seq_sub_spl_rate = configs['seq_sub_spl_rate']
        if "pcr_deletion" in configs.keys():
            pcr_deletion = configs['pcr_deletion']
        if "pcr_insertion" in configs.keys():
            pcr_insertion = configs['pcr_insertion']
        if "pcr_del_rate" in configs.keys():
            pcr_del_rate = configs['pcr_del_rate']
        if "pcr_ins_rate" in configs.keys():
            pcr_ins_rate = configs['pcr_ins_rate']
        if "seq_deletion" in configs.keys():
            seq_deletion = configs['seq_deletion']
        if "seq_insertion" in configs.keys():
            seq_insertion = configs['seq_insertion']
        if "seq_del_rate" in configs.keys():
            seq_del_rate = configs['seq_del_rate']
        if "seq_ins_rate" in configs.keys():
            seq_ins_rate = configs['seq_ins_rate']

    stime = time.time()
    res_dict = umilen(
        len_params=len_params,
        mode=kwargs['mode'] if 'mode' in kwargs.keys() else None,
        material_params=kwargs['material_params'] if 'material_params' in kwargs.keys() else None,
        seq_params=kwargs['seq_params'] if 'seq_params' in kwargs.keys() else None,
        seq_num=seq_num,
        is_sv_umi_lib=is_sv_umi_lib,
        is_sv_seq_lib=is_sv_seq_lib,
        is_sv_primer_lib=is_sv_primer_lib,
        is_sv_adapter_lib=is_sv_adapter_lib,
        is_sv_spacer_lib=is_sv_spacer_lib,
        working_dir=working_dir,
        condis=condis,
        sim_thres=sim_thres,
        permutation=permutation,
        err_route=err_route,
        ampl_rate=ampl_rate,
        pcr_error=pcr_error,
        seq_error=seq_error,
        pcr_num=pcr_num,

        # indels
        pcr_deletion=pcr_deletion,
        pcr_insertion=pcr_insertion,
        pcr_del_rate=pcr_del_rate,
        pcr_ins_rate=pcr_ins_rate,
        seq_deletion=seq_deletion,
        seq_insertion=seq_insertion,
        seq_del_rate=seq_del_rate,
        seq_ins_rate=seq_ins_rate,

        err_num_met=err_num_met,
        use_seed=use_seed,
        seed=seed,
        sv_fastq_fp=sv_fastq_fp,
        seq_sub_spl_number=seq_sub_spl_number,
        seq_sub_spl_rate=seq_sub_spl_rate,
        verbose=verbose,
    ).generate()
    etime = time.time()
    print("===>Time: {:.3f}s".format(etime - stime))
    print('Finished!')
    return res_dict


def simu_seq_dep(
        seq_num,
        working_dir,
        sim_thres,
        permutation,

        len_params=None,
        condis=None,
        ampl_rate=None,
        err_route=None,
        pcr_error=None,
        seq_error=None,
        pcr_num=None,
        err_num_met=None,

        use_seed=True,
        seed=1,

        pcr_deletion=False,
        pcr_insertion=False,
        pcr_del_rate=0,
        pcr_ins_rate=0,
        seq_deletion=False,
        seq_insertion=False,
        seq_del_rate=0,
        seq_ins_rate=0,

        is_sv_umi_lib=True,
        is_sv_seq_lib=True,
        is_sv_primer_lib=True,
        is_sv_adapter_lib=True,
        is_sv_spacer_lib=True,
        sv_fastq_fp=True,

        seq_sub_spl_numbers=None,
        seq_sub_spl_rate=1 / 3,

        config_fpn=None,

        verbose=True,
        **kwargs,
):
    if config_fpn:
        import yaml
        with open(config_fpn, "r") as f:
            configs = yaml.safe_load(f)
            # for k, item in configs.items():
            #     print(k, item)
        len_params = configs['len_params']
        kwargs['seq_params'] = configs['seq_params']
        kwargs['material_params'] = configs['material_params']
        condis = configs['condis']

        ampl_rate = configs['ampl_rate']
        err_route = configs['err_route']
        seq_error = configs['seq_error']
        pcr_num = configs['pcr_num']
        err_num_met = configs['err_num_met']
        pcr_error = configs['pcr_error']
        seq_sub_spl_numbers = configs['seq_sub_spl_numbers']
        seq_sub_spl_rate = configs['seq_sub_spl_rate']
        if "pcr_deletion" in configs.keys():
            pcr_deletion = configs['pcr_deletion']
        if "pcr_insertion" in configs.keys():
            pcr_insertion = configs['pcr_insertion']
        if "pcr_del_rate" in configs.keys():
            pcr_del_rate = configs['pcr_del_rate']
        if "pcr_ins_rate" in configs.keys():
            pcr_ins_rate = configs['pcr_ins_rate']
        if "seq_deletion" in configs.keys():
            seq_deletion = configs['seq_deletion']
        if "seq_insertion" in configs.keys():
            seq_insertion = configs['seq_insertion']
        if "seq_del_rate" in configs.keys():
            seq_del_rate = configs['seq_del_rate']
        if "seq_ins_rate" in configs.keys():
            seq_ins_rate = configs['seq_ins_rate']

    stime = time.time()
    res_dict = seqdep(
        len_params=len_params,
        mode=kwargs['mode'] if 'mode' in kwargs.keys() else None,
        material_params=kwargs['material_params'] if 'material_params' in kwargs.keys() else None,
        seq_params=kwargs['seq_params'] if 'seq_params' in kwargs.keys() else None,
        seq_num=seq_num,
        is_sv_umi_lib=is_sv_umi_lib,
        is_sv_seq_lib=is_sv_seq_lib,
        is_sv_primer_lib=is_sv_primer_lib,
        is_sv_adapter_lib=is_sv_adapter_lib,
        is_sv_spacer_lib=is_sv_spacer_lib,
        working_dir=working_dir,
        condis=condis,
        sim_thres=sim_thres,
        permutation=permutation,
        err_route=err_route,
        pcr_error=pcr_error,
        seq_error=seq_error,
        pcr_num=pcr_num,
        err_num_met=err_num_met,

        # indels
        pcr_deletion=pcr_deletion,
        pcr_insertion=pcr_insertion,
        pcr_del_rate=pcr_del_rate,
        pcr_ins_rate=pcr_ins_rate,
        seq_deletion=seq_deletion,
        seq_insertion=seq_insertion,
        seq_del_rate=seq_del_rate,
        seq_ins_rate=seq_ins_rate,

        use_seed=use_seed,
        seed=seed,
        ampl_rate=ampl_rate,
        sv_fastq_fp=sv_fastq_fp,
        seq_sub_spl_numbers=seq_sub_spl_numbers,
        seq_sub_spl_rate=seq_sub_spl_rate,
        verbose=verbose,
    ).generate()
    etime = time.time()
    print("===>Time: {:.3f}s".format(etime - stime))
    print('Finished!')
    return res_dict


def simu_generic(
        seq_num,
        working_dir,
        sim_thres,
        permutation,

        len_params=None,
        condis=None,
        ampl_rate=None,
        err_route=None,
        pcr_error=None,
        seq_error=None,
        pcr_num=None,
        err_num_met=None,

        use_seed=True,
        seed=1,

        bead_mutation=False,
        bead_mut_rate=1e-4,
        bead_deletion=False,
        bead_del_rate=0.016,
        bead_insertion=False,
        bead_ins_rate=7.1e-7,

        pcr_deletion=False,
        pcr_insertion=False,
        pcr_del_rate=0,
        pcr_ins_rate=0,
        seq_deletion=False,
        seq_insertion=False,
        seq_del_rate=0,
        seq_ins_rate=0,

        is_sv_umi_lib=True,
        is_sv_seq_lib=True,
        is_sv_primer_lib=True,
        is_sv_adapter_lib=True,
        is_sv_spacer_lib=True,
        sv_fastq_fp=True,
        sv_fastq_fn='example',

        seq_sub_spl_number=None,
        seq_sub_spl_rate=1 / 3,

        config_fpn=None,

        verbose=True,
        **kwargs,
):
    if config_fpn:
        import yaml
        with open(config_fpn, "r") as f:
            configs = yaml.safe_load(f)
            # for k, item in configs.items():
            #     print(k, item)
        len_params = configs['len_params']
        kwargs['seq_params'] = configs['seq_params']
        kwargs['material_params'] = configs['material_params']
        condis = configs['condis']

        ampl_rate = configs['ampl_rate']
        err_route = configs['err_route']
        seq_error = configs['seq_error']
        pcr_num = configs['pcr_num']
        err_num_met = configs['err_num_met']
        pcr_error = configs['pcr_error']
        seq_sub_spl_number = configs['seq_sub_spl_number']
        seq_sub_spl_rate = configs['seq_sub_spl_rate']
        sv_fastq_fn = configs['sv_fastq_fn']
        if "bead_mutation" in configs.keys():
            bead_mutation = configs['bead_mutation']
        if "bead_deletion" in configs.keys():
            bead_deletion = configs['bead_deletion']
        if "bead_insertion" in configs.keys():
            bead_insertion = configs['bead_insertion']
        if "bead_mut_rate" in configs.keys():
            bead_mut_rate = configs['bead_mut_rate']
        if "bead_del_rate" in configs.keys():
            bead_del_rate = configs['bead_del_rate']
        if "bead_ins_rate" in configs.keys():
            bead_ins_rate = configs['bead_ins_rate']
        if "pcr_deletion" in configs.keys():
            pcr_deletion = configs['pcr_deletion']
        if "pcr_insertion" in configs.keys():
            pcr_insertion = configs['pcr_insertion']
        if "pcr_del_rate" in configs.keys():
            pcr_del_rate = configs['pcr_del_rate']
        if "pcr_ins_rate" in configs.keys():
            pcr_ins_rate = configs['pcr_ins_rate']
        if "seq_deletion" in configs.keys():
            seq_deletion = configs['seq_deletion']
        if "seq_insertion" in configs.keys():
            seq_insertion = configs['seq_insertion']
        if "seq_del_rate" in configs.keys():
            seq_del_rate = configs['seq_del_rate']
        if "seq_ins_rate" in configs.keys():
            seq_ins_rate = configs['seq_ins_rate']

    stime = time.time()
    res_dict = generic(
        len_params=len_params,
        mode=kwargs['mode'] if 'mode' in kwargs.keys() else None,
        material_params=kwargs['material_params'] if 'material_params' in kwargs.keys() else None,
        seq_params=kwargs['seq_params'] if 'seq_params' in kwargs.keys() else None,
        seq_num=seq_num,
        is_sv_umi_lib=is_sv_umi_lib,
        is_sv_seq_lib=is_sv_seq_lib,
        is_sv_primer_lib=is_sv_primer_lib,
        is_sv_adapter_lib=is_sv_adapter_lib,
        is_sv_spacer_lib=is_sv_spacer_lib,
        working_dir=working_dir,
        condis=condis,
        sim_thres=sim_thres,
        permutation=permutation,
        err_route=err_route,
        pcr_error=pcr_error,
        seq_error=seq_error,
        pcr_num=pcr_num,
        err_num_met=err_num_met,

        bead_mutation=bead_mutation,
        bead_deletion=bead_deletion,
        bead_insertion=bead_insertion,
        bead_mut_rate=bead_mut_rate,
        bead_del_rate=bead_del_rate,
        bead_ins_rate=bead_ins_rate,

        # indels
        pcr_deletion=pcr_deletion,
        pcr_insertion=pcr_insertion,
        pcr_del_rate=pcr_del_rate,
        pcr_ins_rate=pcr_ins_rate,
        seq_deletion=seq_deletion,
        seq_insertion=seq_insertion,
        seq_del_rate=seq_del_rate,
        seq_ins_rate=seq_ins_rate,

        use_seed=use_seed,
        seed=seed,
        ampl_rate=ampl_rate,
        sv_fastq_fp=sv_fastq_fp,
        sv_fastq_fn=sv_fastq_fn,
        seq_sub_spl_number=seq_sub_spl_number,
        seq_sub_spl_rate=seq_sub_spl_rate,
        verbose=verbose,
    ).generate()
    etime = time.time()
    print("===>Time: {:.3f}s".format(etime - stime))
    print('Finished!')
    return res_dict


if __name__ == "__main__":
    from tresor.path import to
    import numpy as np

    # library(
    #     seq_num=50,
    #     len_params={
    #         'umi': {
    #             'umi_unit_pattern': 3,
    #             'umi_unit_len': 12,
    #         },
    #         'umi_1': {
    #             'umi_unit_pattern': 3,
    #             'umi_unit_len': 12,
    #         },
    #         'seq': 100,
    #         'seq_2': 100,
    #         'adapter': 10,
    #         'adapter_1': 10,
    #         'primer': 10,
    #         'primer_1': 10,
    #         'spacer': 10,
    #         'spacer_1': 10,
    #     },
    #     seq_params={
    #         'custom': 'AAGC',  # BAGC
    #         'custom_1': 'V',
    #     },
    #     material_params={
    #         # 'fasta_cdna_fpn': to('data/Homo_sapiens.GRCh38.cdna.all.fa.gz'),  # None False
    #     },
    #     is_seed=True,
    #
    #     working_dir=to('data/simu/docs/'),
    #
    #     # condis=['umi'],
    #     # condis=['umi', 'seq'],
    #     condis=['umi', 'custom', 'seq', 'custom_1'],
    #     # condis=['umi', 'primer', 'primer_1', 'spacer', 'spacer_1', 'adapter', 'adapter_1', 'seq', 'seq_2', 'umi_1'],
    #     sim_thres=3,
    #     permutation=0,
    #
    #     bead_mutation=True,  # True False
    #     bead_mut_rate=1e-4,  # 0.016 0.00004
    #     bead_deletion=True,  # True False
    #     bead_insertion=True,
    #     bead_del_rate=0.016,  # 0.016 0.00004, 2.4e-7
    #     bead_ins_rate=0.1,  # 0.011 0.00001, 7.1e-7
    #
    #     mode='short_read',  # long_read short_read
    #
    #     verbose=True,  # False True
    # )

    # for perm_i in range(1):
    #     simu_seq_err(
    #         len_params={
    #             'umi': {
    #                 'umi_unit_pattern': 1,
    #                 'umi_unit_len': 10,
    #             },
    #             'seq': 100,
    #         },
    #         seq_params={
    #             'custom': 'AAGC',
    #             'custom_1': 'A',
    #         },
    #         material_params={
    #             'fasta_cdna_fpn': to('data/Homo_sapiens.GRCh38.cdna.all.fa.gz'),  # None False
    #         },
    #         seq_num=50,
    #         working_dir=to('data/simu/docs/') + 'permute_' + str(perm_i) + '/',
    #
    #         is_sv_umi_lib=True,
    #         is_sv_seq_lib=True,
    #         is_sv_primer_lib=True,
    #         is_sv_adapter_lib=True,
    #         is_sv_spacer_lib=True,
    #         condis=['umi'],
    #         # condis=['umi', 'seq'],
    #         # condis=['umi', 'custom', 'seq', 'custom_1'],
    #         sim_thres=3,
    #         permutation=perm_i, # 0
    #
    #         # PCR amplification
    #         ampl_rate=0.85,
    #         err_route='sptree',  # bftree sptree err1d err2d mutation_table_minimum mutation_table_complete
    #         pcr_error=1e-5,
    #         pcr_num=8,
    #         err_num_met='nbinomial',
    #         seq_errors=[
    #             1e-05,
    #             2.5e-05,
    #             5e-05,
    #             7.5e-05,
    #             0.0001,
    #             0.00025,
    #             0.0005,
    #             0.00075,
    #             0.001,
    #             0.0025,
    #             0.005,
    #             0.0075,
    #             0.01,
    #             0.025,
    #             0.05,
    #             0.075,
    #             0.1,
    #             0.2,
    #             0.3,
    #         ],
    #         seq_sub_spl_number=None,  # 200 or None
    #         seq_sub_spl_rate=1, # 0.333 or 1
    #         use_seed=True,
    #         seed=1,
    #
    #         mode='short_read',  # long_read short_read
    #         sv_fastq_fp=to('data/simu/docs/') + 'permute_' + str(perm_i) + '/',
    #
    #         verbose=False,  # True
    #     )

    # for perm_i in range(1):
    #     print(perm_i)
    #     simu_pcr_err(
    #         # initial sequence generation
    #         len_params={
    #             'umi': {
    #                 'umi_unit_pattern': 1,
    #                 'umi_unit_len': 10,
    #             },
    #             'umi_1': {
    #                 'umi_unit_pattern': 3,
    #                 'umi_unit_len': 12,
    #             },
    #             'barcode': 16,
    #             'seq': 100,
    #             'seq_2': 100,
    #             'adapter': 10,
    #             'adapter_1': 10,
    #             'primer': 10,
    #             'primer_1': 10,
    #             'spacer': 10,
    #             'spacer_1': 10,
    #         },
    #         seq_params={
    #             'custom': 'AAGC',
    #             'custom_1': 'A',
    #         },
    #         material_params={
    #             'fasta_cdna_fpn': to('data/Homo_sapiens.GRCh38.cdna.all.fa.gz'),  # None False
    #         },
    #         seq_num=50,
    #         # working_dir=to('data/simu/'),
    #         working_dir=to('data/simu/docs/') + 'permute_' + str(perm_i) + '/',
    #
    #         is_sv_umi_lib=True,
    #         is_sv_seq_lib=True,
    #         is_sv_primer_lib=True,
    #         is_sv_adapter_lib=True,
    #         is_sv_spacer_lib=True,
    #         condis=['umi'],
    #         # condis=['umi', 'seq'],
    #         # condis=['umi', 'custom', 'seq', 'custom_1'],
    #         sim_thres=3,
    #         permutation=perm_i, # 0
    #
    #         # PCR amplification
    #         ampl_rate=0.85,
    #         err_route='sptree',  # bftree sptree err1d err2d mutation_table_minimum mutation_table_complete
    #         pcr_errors=[1e-05, 2.5e-05, 5e-05, 7.5e-05, 0.0001, 0.00025, 0.0005, 0.00075, 0.001, 0.0025, 0.005, 0.0075,
    #                     0.01, 0.05],
    #         pcr_num=8,
    #         err_num_met='nbinomial',
    #         seq_error=0.001,
    #         # seq_sub_spl_number=200,  # None
    #         seq_sub_spl_rate=1, # 0.333
    #         use_seed=True,
    #         seed=1,
    #
    #         verbose=False,  # True False
    #         mode='short_read',  # long_read short_read
    #         # sv_fastq_fp=to('data/simu/'),
    #         sv_fastq_fp=to('data/simu/docs/') + 'permute_' + str(perm_i) + '/',
    #     )

    # for perm_i in range(1):
    #     print(perm_i)
    #     simu_pcr_num(
    #         len_params={
    #             'umi': {
    #                 'umi_unit_pattern': 1,
    #                 'umi_unit_len': 10,
    #             },
    #             'umi_1': {
    #                 'umi_unit_pattern': 3,
    #                 'umi_unit_len': 12,
    #             },
    #             'barcode': 16,
    #             'seq': 100,
    #             'seq_2': 100,
    #             'adapter': 10,
    #             'adapter_1': 10,
    #             'primer': 10,
    #             'primer_1': 10,
    #             'spacer': 10,
    #             'spacer_1': 10,
    #         },
    #         seq_params={
    #             'custom': 'AAGC',
    #             'custom_1': 'A',
    #         },
    #         material_params={
    #             'fasta_cdna_fpn': to('data/Homo_sapiens.GRCh38.cdna.all.fa.gz'),  # None False
    #         },
    #         seq_num=50,
    #         working_dir=to('data/simu/docs/') + 'permute_' + str(perm_i) + '/',
    #         is_sv_umi_lib=True,
    #         is_sv_seq_lib=True,
    #         is_sv_primer_lib=True,
    #         is_sv_adapter_lib=True,
    #         is_sv_spacer_lib=True,
    #         condis=['umi'],
    #         # condis=['umi', 'seq'],
    #         # condis=['umi', 'custom', 'seq', 'custom_1'],
    #         sim_thres=3,
    #         permutation=perm_i, # 0
    #
    #         # PCR amplification
    #         ampl_rate=0.85,
    #         err_route='sptree',  # bftree sptree err1d err2d mutation_table_minimum mutation_table_complete
    #         pcr_error=1e-4,
    #         pcr_nums=np.arange(1, 16 + 1, 1),
    #         err_num_met='nbinomial',
    #         seq_error=0.01,
    #         seq_sub_spl_number=None,  # None 200
    #         seq_sub_spl_rate=0.333, # 0.333 1
    #         use_seed=True,
    #         seed=1,
    #
    #         verbose=False,  # True False
    #         mode='short_read',  # long_read short_read
    #         sv_fastq_fp=to('data/simu/docs/') + 'permute_' + str(perm_i) + '/',
    #     )

    # for perm_i in range(1):
    #     print(perm_i)
    #     simu_ampl_rate(
    #         # initial sequence generation
    #
    #         len_params={
    #             'umi': {
    #                 'umi_unit_pattern': 1,
    #                 'umi_unit_len': 10,
    #             },
    #             'umi_1': {
    #                 'umi_unit_pattern': 3,
    #                 'umi_unit_len': 12,
    #             },
    #             'barcode': 16,
    #             'seq': 100,
    #             'seq_2': 100,
    #             'adapter': 10,
    #             'adapter_1': 10,
    #             'primer': 10,
    #             'primer_1': 10,
    #             'spacer': 10,
    #             'spacer_1': 10,
    #         },
    #         seq_params={
    #             'custom': 'AAGC',
    #             'custom_1': 'A',
    #         },
    #         material_params={
    #             'fasta_cdna_fpn': to('data/Homo_sapiens.GRCh38.cdna.all.fa.gz'),  # None False
    #         },
    #         seq_num=50,
    #         # working_dir=to('data/simu/'),
    #         working_dir=to('data/simu/docs/') + 'permute_' + str(perm_i) + '/',
    #
    #         is_sv_umi_lib=True,
    #         is_sv_seq_lib=True,
    #         is_sv_primer_lib=True,
    #         is_sv_adapter_lib=True,
    #         is_sv_spacer_lib=True,
    #         # condis=['umi'],
    #         condis=['umi', 'seq'],
    #         # condis=['umi', 'custom', 'seq', 'custom_1'],
    #         sim_thres=3,
    #         permutation=perm_i,
    #
    #         # PCR amplification
    #         ampl_rates=np.linspace(0.1, 1, 10),
    #         err_route='err2d',  # bftree sptree err1d err2d mutation_table_minimum mutation_table_complete
    #         pcr_error=1e-5,
    #         pcr_num=8,
    #         err_num_met='nbinomial', # nbinomial
    #         seq_error=0.001,
    #         # seq_sub_spl_number=200, # None
    #         seq_sub_spl_rate=1,  # 0.333
    #         use_seed=True,
    #         seed=1,
    #
    #         verbose=True,  # True False
    #         mode='short_read',  # long_read short_read
    #         # sv_fastq_fp=to('data/simu/'),
    #         sv_fastq_fp=to('data/simu/docs/') + 'permute_' + str(perm_i) + '/',
    #     )

    # for perm_i in range(1):
    #     print(perm_i)
    #     simu_umi_len(
    #         # initial sequence generation
    #         len_params={
    #             'umi': {
    #                 'umi_unit_pattern': 1,
    #                 'umi_unit_lens': np.arange(6, 18 + 1, 1),
    #             },
    #             'umi_1': {
    #                 'umi_unit_pattern': 3,
    #                 'umi_unit_len': 12,
    #             },
    #             'barcode': 16,
    #             'seq': 100,
    #             'seq_2': 100,
    #             'adapter': 10,
    #             'adapter_1': 10,
    #             'primer': 10,
    #             'primer_1': 10,
    #             'spacer': 10,
    #             'spacer_1': 10,
    #         },
    #         seq_params={
    #             'custom': 'AAGC',
    #             'custom_1': 'A',
    #         },
    #         material_params={
    #             'fasta_cdna_fpn': to('data/Homo_sapiens.GRCh38.cdna.all.fa.gz'),  # None False
    #         },
    #         seq_num=50,
    #         # working_dir=to('data/simu/'),
    #         working_dir=to('data/simu/') + 'permute_' + str(perm_i) + '/',
    #
    #         is_sv_umi_lib=True,
    #         is_sv_seq_lib=True,
    #         is_sv_primer_lib=True,
    #         is_sv_adapter_lib=True,
    #         is_sv_spacer_lib=True,
    #         # condis=['umi'],
    #         condis=['umi', 'seq'],
    #         # condis=['umi', 'custom', 'seq', 'custom_1'],
    #         sim_thres=3,
    #         permutation=perm_i, # 0
    #
    #         # PCR amplification
    #         ampl_rate=0.85,
    #         err_route='sptree',  # bftree sptree err1d err2d mutation_table_minimum mutation_table_complete
    #         pcr_error=1e-5,
    #         pcr_num=8,
    #         err_num_met='nbinomial',
    #         seq_error=0.001,
    #         seq_sub_spl_number=None,  # None 200
    #         seq_sub_spl_rate=1, # 0.333
    #         use_seed=True,
    #         seed=1,
    #
    #         verbose=False,  # True False
    #         mode='short_read',  # long_read short_read
    #         # sv_fastq_fp=to('data/simu/'),
    #         sv_fastq_fp=to('data/simu/') + 'permute_' + str(perm_i) + '/',
    #     )

    # for perm_i in range(1):
    #     print(perm_i)
    #     simu_seq_dep(
    #         # initial sequence generation
    #         len_params={
    #             'umi': {
    #                 'umi_unit_pattern': 1,
    #                 'umi_unit_len': 10,
    #             },
    #             'umi_1': {
    #                 'umi_unit_pattern': 3,
    #                 'umi_unit_len': 12,
    #             },
    #             'barcode': 16,
    #             'seq': 100,
    #             'seq_2': 100,
    #             'adapter': 10,
    #             'adapter_1': 10,
    #             'primer': 10,
    #             'primer_1': 10,
    #             'spacer': 10,
    #             'spacer_1': 10,
    #         },
    #         seq_params={
    #             'custom': 'AAGC',
    #             'custom_1': 'A',
    #         },
    #         material_params={
    #             'fasta_cdna_fpn': to('data/Homo_sapiens.GRCh38.cdna.all.fa.gz'),  # None False
    #         },
    #         seq_num=50,
    #         # working_dir=to('data/simu/'),
    #         working_dir=to('data/simu/docs/') + 'permute_' + str(perm_i) + '/',
    #
    #         is_sv_umi_lib=True,
    #         is_sv_seq_lib=True,
    #         is_sv_primer_lib=True,
    #         is_sv_adapter_lib=True,
    #         is_sv_spacer_lib=True,
    #         condis=['umi'],
    #         # condis=['umi', 'seq'],
    #         # condis=['umi', 'custom', 'seq', 'custom_1'],
    #         sim_thres=3,
    #         permutation=perm_i, # 0
    #
    #         # PCR amplification
    #         ampl_rate=0.85,
    #         err_route='err2d',  # bftree sptree err1d err2d mutation_table_minimum mutation_table_complete
    #         pcr_error=1e-05,
    #         pcr_num=8,
    #         err_num_met='nbinomial',
    #         seq_error=0.001,
    #         seq_sub_spl_numbers=[100, 200, 500, 600, 800, 1000, 2000, 3000, 5000],  # None 200
    #         # seq_sub_spl_rate=0.333,
    #         use_seed=True,
    #         seed=1,
    #
    #         verbose=False,  # True False
    #         mode='short_read',  # long_read short_read
    #         # sv_fastq_fp=to('data/simu/'),
    #         sv_fastq_fp=to('data/simu/docs/') + 'permute_' + str(perm_i) + '/',
    #     )

    # @@
    for i, umi_num_i in enumerate([
        1,
        # 2, 3, 4, 5, 6, 7, 8, 9, 10,
        # 11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
        # 21, 22, 23, 24, 25, 26, 27, 28, 29, 30,
        # 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
        # 41, 42, 43, 44, 45,
    ]):
        simu_generic(
            # initial sequence generation
            len_params={
                'umi': {
                    'umi_unit_pattern': 1,
                    'umi_unit_len': 10,
                },
                'umi_1': {
                    'umi_unit_pattern': 3,
                    'umi_unit_len': 12,
                },
                'barcode': 16,
                'seq': 100,
                'seq_2': 100,
                'adapter': 10,
                'adapter_1': 10,
                'primer': 10,
                'primer_1': 10,
                'spacer': 10,
                'spacer_1': 10,
            },
            seq_params={
                'custom': 'AAGC',
                'custom_1': 'A',
            },
            material_params={
                'fasta_cdna_fpn': to('data/Homo_sapiens.GRCh38.cdna.all.fa.gz'),  # None False
            },
            seq_num=umi_num_i,
            working_dir=to('data/simu/'),

            is_sv_umi_lib=True,
            is_sv_seq_lib=True,
            is_sv_primer_lib=True,
            is_sv_adapter_lib=True,
            is_sv_spacer_lib=True,
            condis=['umi'],
            # condis=['umi', 'seq'],
            # condis=['umi', 'custom', 'seq', 'custom_1'],
            sim_thres=3,
            permutation=0,

            # PCR amplification
            ampl_rate=0.85,
            err_route='err2d',  # bftree sptree err1d err2d mutation_table_minimum mutation_table_complete
            pcr_error=1e-05,
            pcr_num=8,
            err_num_met='nbinomial',
            seq_error=0.05, # 0.005 0.01 0.05 0.1
            seq_sub_spl_number=None, # None 200 2000
            seq_sub_spl_rate=1, # 0.333
            use_seed=True,
            seed=1,

            verbose=False,  # True False

            mode='short_read',  # long_read short_read

            sv_fastq_fp=to('data/simu/'),
            sv_fastq_fn='umi_num_' +  str(i),
        )


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


from tresor.read.Library import Library as liblogginger
from tresor.read.inf.Pseudo import Pseudo as seqpseudo


class Design(seqpseudo):

    def __init__(self, *args, **kwargs):
        super(Design, self).__init__(*args, **kwargs)
        self.args = args
        self.kwargs = kwargs

    @liblogginger(method='default')
    def general(self, **kwargs):
        return ''.join([
            self.kwargs['dna_map'][i] for i in
            self.kwargs['pseudorandom_num']
        ])

    @liblogginger(method='default')
    def reoccur(self, **kwargs):
        return ''.join([
            self.kwargs['dna_map'][i] * self.kwargs['umi_unit_pattern'] for i in
            self.kwargs['pseudorandom_num']
        ])

    @liblogginger(method='separate')
    def write(self, **kwargs):
        return 'written'


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"


import time
import numpy as np
import pandas as pd
from functools import wraps
from tresor.util.Console import Console


class Ordering:

    def __init__(
            self,
            method,
    ):
        self.method = method
        self.console = Console()

    def __call__(self, deal):
        if self.method == 'shuffle':
            order = self.shuffle
        elif self.method == 'permute':
            order = self.permute
        else:
            order = self.shuffle
        @wraps(deal)
        def switch(dself, *args, **kwargs):
            # print(args)
            # print(kwargs)
            if 'verbose' in kwargs['params'].keys():
                self.console.verbose = kwargs['params']['verbose']
            else:
                self.console.verbose = True
            self.console.print('======>ordering...')
            res2p = deal(dself, **kwargs)
            # print(res2p['data'].shape[0])
            res2p['data'] = order(
                data=res2p['data'],
                # use_seed=kwargs['data'],
                # seed=kwargs['data'],
            )
            # print(res2p['data'])
            return res2p
        return switch

    def shuffle(self, data, use_seed=True, seed=1):
        num_samples = len(data)
        self.console.print('=========>The number of objects to be shuffling: {}'.format(num_samples))
        if isinstance(data, pd.DataFrame):
            if use_seed:
                state = np.random.RandomState(seed)
                data = data.iloc[state.shuffle(num_samples)].reset_index(drop=True)
                self.console.print(data)
            else:
                data = data.iloc[np.random.shuffle(num_samples)].reset_index(drop=True)
        elif type(data) is np.ndarray:
            if use_seed:
                self.console.print('=========>start shuffling...')
                stime = time.time()
                state = np.random.RandomState(seed)
                state.shuffle(data)
                self.console.print('=========>shuffling time: {}'.format(time.time() - stime))
            else:
                np.random.shuffle(data)
        else:
            if use_seed:
                state = np.random.RandomState(seed)
                ids = state.shuffle(data)
                data = [data[i - 1] for i in ids]
            else:
                ids = np.random.shuffle(data)
                data = [data[i - 1] for i in ids]
        return data

    def permute(self, ):
        pass


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"



class Hamming:

    def __init__(self, ):
        pass

    def general(self, s1, s2):
        return sum(i != j for i, j in zip(s1, s2))


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


import gzip
from Bio import SeqIO


class Fasta:

    def __init__(self):
        pass

    def get_from_gz(self, fasta_fpn):
        arr = []
        with gzip.open(fasta_fpn, "rt") as handle:
            for record in SeqIO.parse(handle, "fasta"):
                arr.append([str(record.id), str(record.seq)])
            # print(arr)
        return arr

    def save_(self, list_2d, sv_fp):
        for i, e in enumerate(list_2d):
            prot_name = str(e[0])
            seq = str(e[1])
            print('No.{} saving {} in FASTA format.'.format(i+1, prot_name))
            f = open(sv_fp + prot_name + '.fasta', 'w')
            f.write('>' + prot_name + '\n')
            f.write(seq + '\n')
            f.close()
        return 0


__version__ = "0.0.1"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


import time
import numpy as np
import pandas as pd

from tresor.library.SingleLocus import SingleLocus as simuip
from tresor.pcr.Amplify import Amplify as pcr
from tresor.sequencing.Calling import Calling as seq
from tresor.pcr.Subsampling import Subsampling
from tresor.util.sequence.fastq.Write import write as wfastq
from tresor.util.file.Writer import Writer as fwriter
from tresor.util.Console import Console


class SingleLocus:

    def __init__(
            self,
            len_params,
            seq_num,
            is_sv_umi_lib,
            is_sv_seq_lib,
            is_sv_primer_lib,
            is_sv_adapter_lib,
            is_sv_spacer_lib,
            working_dir,
            condis,
            sim_thres,
            permutation,

            seq_error,
            err_route,
            pcr_error,
            pcr_num,
            err_num_met,
            use_seed,
            seed,

            ampl_rate,
            sv_fastq_fp,
            sv_fastq_fn,

            seq_sub_spl_number=None,
            seq_sub_spl_rate=1/3,

            verbose=True,
            **kwargs,
    ):
        self.len_params = len_params
        self.seq_num = seq_num
        self.is_sv_umi_lib = is_sv_umi_lib
        self.is_sv_seq_lib = is_sv_seq_lib
        self.is_sv_primer_lib = is_sv_primer_lib
        self.is_sv_adapter_lib = is_sv_adapter_lib
        self.is_sv_spacer_lib = is_sv_spacer_lib

        self.working_dir = working_dir
        self.condis = condis
        self.sim_thres = sim_thres
        self.permutation = permutation

        # self.bead_mutation = kwargs['bead_mutation']
        # self.bead_mut_rate = kwargs['bead_mut_rate']
        # self.bead_deletion = kwargs['bead_deletion']
        # self.bead_del_rate = kwargs['bead_del_rate']
        # self.bead_insertion = kwargs['bead_insertion']
        # self.bead_ins_rate = kwargs['bead_ins_rate']

        self.err_route = err_route
        self.ampl_rate = ampl_rate
        self.pcr_num = pcr_num
        self.pcr_error = pcr_error
        self.seq_error = seq_error
        self.err_num_met = err_num_met
        self.use_seed = use_seed
        self.seed = seed

        self.seq_sub_spl_number = seq_sub_spl_number
        self.seq_sub_spl_rate = seq_sub_spl_rate
        self.sv_fastq_fp = sv_fastq_fp
        self.sv_fastq_fn = sv_fastq_fn

        self.pcr = pcr
        self.seq = seq
        self.wfastq = wfastq
        self.subsampling = Subsampling(verbose=verbose)
        self.fwriter = fwriter()

        self.kwargs = kwargs
        print(self.kwargs)

        self.console = Console()
        self.console.verbose = verbose
        self.verbose = verbose

        ### +++++++++++++++ block: generate sequencing library +++++++++++++++
        self.console.print('===>Sequencing library generation starts')
        self.sequencing_library, self.lib_err_mark = simuip(
            len_params=self.len_params,
            seq_num=self.seq_num,
            is_seed=self.use_seed,
            working_dir=self.working_dir,
            condis=self.condis,
            sim_thres=self.sim_thres,
            # bead_mutation=self.bead_mutation,
            # bead_mut_rate=self.bead_mut_rate,
            # bead_deletion=self.bead_deletion,
            # bead_del_rate=self.bead_del_rate,
            # bead_insertion=self.bead_insertion,
            bead_mutation=self.kwargs['bead_mutation'] if 'bead_mutation' in self.kwargs.keys() else False,
            bead_mut_rate=self.kwargs['bead_mut_rate'] if 'bead_mut_rate' in self.kwargs.keys() else False,
            bead_deletion=self.kwargs['bead_deletion'] if 'bead_deletion' in self.kwargs.keys() else False,
            bead_del_rate=self.kwargs['bead_del_rate'] if 'bead_del_rate' in self.kwargs.keys() else False,
            bead_insertion=self.kwargs['bead_insertion'] if 'bead_insertion' in self.kwargs.keys() else False,
            bead_ins_rate=self.kwargs['bead_del_rate'] if 'bead_del_rate' in self.kwargs.keys() else False,
            permutation=self.permutation,
            is_sv_umi_lib=self.is_sv_umi_lib,
            is_sv_seq_lib=self.is_sv_seq_lib,
            is_sv_primer_lib=self.is_sv_primer_lib,
            is_sv_adapter_lib=self.is_sv_adapter_lib,
            is_sv_spacer_lib=self.is_sv_spacer_lib,
            verbose=self.verbose,
            mode=self.kwargs['mode'],
            material_params=self.kwargs['material_params'] if 'material_params' in self.kwargs.keys() else None,
            seq_params=self.kwargs['seq_params'] if 'seq_params' in self.kwargs.keys() else None,
        ).pooling()
        self.lib_err_mark['pcr_err_mark'] = False
        # print(self.lib_err_mark)
        self.sequencing_library = pd.concat([pd.DataFrame(self.sequencing_library), self.lib_err_mark], axis=1).values
        # print(self.sequencing_library)
        self.console.print('===>Sequencing library has been generated')
        # print(self.lib_err_mark)

    def generate(self, ):
        """

        [['1' '1' '0']
         ['1' '1' '1']
         ['1' '1' '2']
         ...
         ['1' '1' '48']
         ['1' '1' '49']]
        Returns
        -------

        """
        ### +++++++++++++++ block: PCR amplification: Preparation +++++++++++++++
        time_arr = []
        satime = time.time()

        self.console.print('===>PCR amplification starts...')
        self.console.print('======>Assign parameters...')
        # print(np.array(self.sequencing_library))

        self.console.print('======>The criterion: {}'.format('12'))
        pcr_ampl_params = {
            'read_lib_fpn': self.working_dir + 'sequencing_library.txt',

            'data': np.array(self.sequencing_library),
            'ampl_rate': self.ampl_rate,
            'pcr_error': self.pcr_error,
            'pcr_num': self.pcr_num,

            'err_route': self.err_route,
            'err_num_met': self.err_num_met,
            'use_seed': self.use_seed,
            'seed': self.seed,
            'recorder_nucleotide_num': [],
            'recorder_pcr_err_num': [],
            'recorder_pcr_read_num': [],

            'seq_sub_spl_number': self.seq_sub_spl_number,
            'seq_sub_spl_rate': self.seq_sub_spl_rate,

            # 'pcr_deletion': False,
            # 'pcr_insertion': False, # False True
            # 'pcr_del_rate': 2.4*10e-6,
            # 'pcr_ins_rate': 7.1*10e-7,

            'pcr_deletion': self.kwargs['pcr_deletion'] if 'pcr_deletion' in self.kwargs.keys() else False,
            'pcr_insertion': self.kwargs['pcr_insertion'] if 'pcr_insertion' in self.kwargs.keys() else False,
            'pcr_del_rate': self.kwargs['pcr_del_rate'] if 'pcr_del_rate' in self.kwargs.keys() else 0,
            'pcr_ins_rate': self.kwargs['pcr_ins_rate'] if 'pcr_ins_rate' in self.kwargs.keys() else 0,

            'verbose': self.verbose,
        }
        # print(pcr_ampl_params['data'])
        # if checking bead synthesis errors, you will get the pcr_ampl_params['data'] like this
        # [['GAT...A' '0' 'init' False False False False]
        #  ['GTA...G' '1' 'init' False False False False]
        #  ['ATG...T' '2' 'init' False False True False]
        #  ['CAA...T' '3' 'init' False False False False]
        # ...
        #  ['TTA...C' '49' 'init' False False False False]]
        # in the normal state, pcr_ampl_params['data'] looks like:
        # pcr_ampl_params['data']
        # [['GGGAAATTTAAACCCTTTAAAGGGAAAAAAGGGCCC' '0' 'init']
        #  ['GGGTTTAAACCCCCCCCCGGGAAATTTTTTGGGTTT' '1' 'init']
        #  ['AAATTTGGGCCCGGGAAAGGGCCCAAAAAAGGGAAA' '2' 'init']
        #  ...
        #  ['CCCAAAAAAGGGCCCAAAGGGCCCTTTGGGGGGCCC' '48' 'init']
        #  ['TTTTTTAAATTTAAAAAAGGGAAAGGGGGGGGGCCC' '49' 'init']]
        # print(pcr_ampl_params['data'][:, 1:3])
        # print(pcr_ampl_params)
        if pcr_ampl_params['err_route'] == 'bftree':
            pcr_ampl_params['data'] = pcr_ampl_params['data'][:, 1:3]
        if pcr_ampl_params['err_route'] == 'sptree':
            pcr_ampl_params['data'] = pcr_ampl_params['data'][:, 1:3]
        if pcr_ampl_params['err_route'] == 'mutation_table_minimum' or pcr_ampl_params['err_route'] == 'mutation_table_complete':
            # print(pcr_ampl_params['data'][:, 0])
            def calc_len(a):
                return len(a)

            vfunc = np.vectorize(calc_len)
            # [[36] vfunc(pcr_ampl_params['data'][:, 0])[:, np.newaxis]
            #  [36]
            #  [36]
            #  ...
            #  [36]
            #  [36]]
            pcr_ampl_params['data'] = np.hstack((
                vfunc(pcr_ampl_params['data'][:, 0])[:, np.newaxis],
                pcr_ampl_params['data'][:, 1:7],
            ))
            # print(pcr_ampl_params)
            # print(pcr_ampl_params['data'])
            # pcr_ampl_params['data']
            # [['36' '0' 'init']
            #  ['36' '1' 'init']
            #  ['36' '2' 'init']
            #  ...
            #  ['36' '48' 'init']
            #  ['36' '49' 'init']]
            col_0 = np.array([[1] for _ in range(pcr_ampl_params['data'].shape[0])])
            mut_info_table = np.hstack((col_0, col_0))
            col_2 = pcr_ampl_params['data'][:, 1].astype(str)[:, np.newaxis]
            mut_info_table = np.hstack((mut_info_table, col_2))

            mut_info_table = pd.concat([pd.DataFrame(mut_info_table), self.lib_err_mark], axis=1).values

            pcr_ampl_params['mut_info'] = mut_info_table
            # print(mut_info_table)
            # print(self.lib_err_mark.values)
            # pcr_ampl_params['mut_info'] = np.empty(shape=[0, 3])
            # print(pcr_ampl_params['mut_info'])

        ### +++++++++++++++ block: PCR amplification: simulation +++++++++++++++
        pcr_stime = time.time()
        pcr = self.pcr(pcr_params=pcr_ampl_params).np()
        # print(pcr.keys())
        self.console.print('======>PCR amplification completes in {}s'.format(time.time() - pcr_stime))
        # print(pcr['data'])
        # print(pcr['data'].shape)

        ### +++++++++++++++ block: Subsampling: sequencing depth or rate +++++++++++++++
        # print(pcr['data'])
        # print(pcr['data'].shape)
        if pcr_ampl_params['err_route'] == 'bftree':
            pcr['data'] = self.subsampling.bftree(pcr_dict=pcr)
        # print(pcr['data'])
        # print(pcr['data'].shape)

        if pcr_ampl_params['err_route'] == 'sptree':
            pcr['data'] = self.subsampling.sptree(pcr_dict=pcr)

        if pcr_ampl_params['err_route'] == 'mutation_table_minimum':
            pcr['data'] = self.subsampling.mutation_table_minimum(pcr_dict=pcr)

        if pcr_ampl_params['err_route'] == 'mutation_table_complete':
            pcr['data'] = self.subsampling.mutation_table_complete(pcr_dict=pcr)
            # print(pcr['data'])
            # print(pcr['data'].shape)

        ### +++++++++++++++ block: Sequencing: parameters +++++++++++++++
        self.console.print('======>Sequencing starts')
        seq_params = {
            'data': pcr['data'],

            'seq_error': self.seq_error,
            'err_num_met': self.err_num_met,
            'use_seed': self.use_seed,
            'seed': self.seed,
            'verbose': self.verbose,

            'seq_sub_spl_number': self.seq_sub_spl_number,
            'seq_sub_spl_rate': self.seq_sub_spl_rate,

            # 'seq_deletion': False,
            # 'seq_insertion': False, # False True
            # 'seq_del_rate': 2.4 * 10e-6,
            # 'seq_ins_rate': 7.1 * 10e-7,

            'seq_deletion': self.kwargs['seq_deletion'] if 'seq_deletion' in self.kwargs.keys() else False,
            'seq_insertion': self.kwargs['seq_insertion'] if 'seq_insertion' in self.kwargs.keys() else False,
            'seq_del_rate': self.kwargs['seq_del_rate'] if 'seq_del_rate' in self.kwargs.keys() else 0,
            'seq_ins_rate': self.kwargs['seq_ins_rate'] if 'seq_ins_rate' in self.kwargs.keys() else 0,
        }
        seq = self.seq(seq_params=seq_params).np()
        self.console.print('=========>Sequencing has completed')
        print('======>simulation completes in {}s'.format(time.time() - satime))
        time_arr.append(time.time() - satime)

        self.console.print('=========>Reads write to files in FastQ format')
        self.wfastq().togz(
            list_2d=seq['data'],
            sv_fp=self.sv_fastq_fp,
            fn=self.sv_fastq_fn,
            symbol='-',
        )
        del seq

        self.console.print('=========>FastQ file is saved')
        self.console.print('======>Simulation completes')
        return {
            'time_arr': time_arr,
        }


if __name__ == "__main__":
    from tresor.path import to

    p = SingleLocus(
        # initial sequence generation
        len_params={
            'umi': {
                'umi_unit_pattern': 1,
                'umi_unit_len': 12,
            },
            'seq': 100,
        },
        seq_params={
            'custom': 'AAGC',
            'custom_1': 'A',
        },
        material_params={
            # 'fasta_cdna_fpn': to('data/Homo_sapiens.GRCh38.cdna.all.fa.gz'),  # None False
        },
        seq_num=50,
        working_dir=to('data/simu/'),

        is_sv_umi_lib=True,
        is_sv_seq_lib=True,
        is_sv_primer_lib=True,
        is_sv_adapter_lib=True,
        is_sv_spacer_lib=True,
        # condis=['umi'],
        condis=['umi', 'seq'],
        # condis=['umi', 'custom', 'seq', 'custom_1'],
        sim_thres=3,
        permutation=0,

        # PCR amplification
        ampl_rate=0.85,
        err_route='mutation_table_complete', # bftree sptree err1d err2d mutation_table_minimum mutation_table_complete
        pcr_error=1e-4,
        pcr_num=10,
        err_num_met='nbinodmial',
        seq_error=0.01,
        seq_sub_spl_number=500, # None 200
        seq_sub_spl_rate=False, # 0.333
        use_seed=True,
        seed=1,

        bead_mutation=True,  # True False
        bead_mut_rate=1e-4,  # 0.016 0.00004
        bead_deletion=True,  # True False
        bead_insertion=True,
        bead_del_rate=0.1/112,  # 0.016 0.00004, 2.4e-7
        bead_ins_rate=7.1e-7,  # 0.011 0.00001, 7.1e-7

        pcr_deletion=True,  # True False
        pcr_insertion=True,
        pcr_del_rate=2.4e-6,  # 0.016 0.00004
        pcr_ins_rate=7.1e-7,  # 0.011 0.00001
        seq_deletion=False,
        seq_insertion=False,
        seq_del_rate=2.4e-6,
        seq_ins_rate=7.1e-7,

        verbose=True, # True False

        mode='short_read',  # long_read short_read

        sv_fastq_fp=to('data/simu/'),
        sv_fastq_fn='example',
    )
    print(p.generate())